/* This file is part of cudaTV.  See the file COPYING for further details. */
#include <stdio.h>

#include "mex.h"
#include "referenceChambolle.h"
#include "chambolleTV.h"
#include "laplacian.h"
#include "primaldualTV.h"


void mexFunction(int nargout, mxArray *argout[], int nargin, const mxArray *argin[]) 
{

	float *input, *cuda_mem;  
	int N;
	size_t size;

	if (nargin<2)
		mexErrMsgTxt("two parameters required, image and itereations");

	//  argout[0] = mxDuplicateArray(argin[0]);
	//  input = (float *)mxGetPr(argout[0]); 
	//  N = mxGetN(argout[0]) * mxGetM(argout[0]);
	//  cudafunc(input,N);

	argout[0] = mxDuplicateArray(argin[0]);
	input = (float *)mxGetPr(argout[0]); 
	int ny = mxGetN(argout[0]);
	int nx = mxGetM(argout[0]);

	int iter=0;
	double* it = (double*) mxGetPr(argin[1]);
	iter = *it;

	int v;
	if (nargin==3) {	
		double* val = (double*) mxGetPr(argin[2]);
		v=(int)( *val );
	} else { v=4; }

	if(v==1){
		mexPrintf("reference implementation grad div\n");
		reflap(input,iter,nx,ny);
	}else if(v==2){
		mexPrintf("reference implementation laplacian stencil\n");
		reflap2(input,iter,nx,ny);
	}else if(v==3){
		mexPrintf("cuda implementation grad div\n");
		cudalap(input,iter,nx,ny);
	}else if(v==4){
		mexPrintf("cuda implementation laplacian stencil\n");
		cudalap2(input,iter,nx,ny);
	}else if(v==5){
		mexPrintf("reference implementation TV chambolle\n");
		float lambda=(float) 10.0;
		refTVcha(input,iter,nx,ny,lambda);
	}else if(v==6){
		mexPrintf("cuda TV chambolle\n");
		float lambda=(float) 10.0;
		cudaTVcha(input,iter,nx,ny,lambda);
	}else if(v==7){
		mexPrintf("cuda TV chambolle 2\n");
		float lambda=(float) 10.0;
		cudaTVcha2(input,iter,nx,ny,lambda);
	}else if(v==11){
		mexPrintf("cuda TV chambolle 2 with structure of arrays (FASTER)\n");
		float lambda=(float) 10.0;
		cudaTVcha2_structure_of_arrays(input,iter,nx,ny,lambda);
	}else if(v==12){
		mexPrintf("cuda TV chambolle 2 dummy\n");
		float lambda=(float) 10.0;
		cudaTVcha2_dummy(input,iter,nx,ny,lambda);
	}else if(v==13){
		mexPrintf("cuda TV primal dual\n");
		float lambda=(float) 10.0;
		cudaTVpd(input,iter,nx,ny,lambda);
	}else if(v==14){
		mexPrintf("cuda TV primal dual with structure of arrays\n");
		float lambda=(float) 10.0;
		cudaTVpd_SoA(input,iter,nx,ny,lambda);
	}else {
		mexPrintf("cuda TV primal dual : PDHG (zhu-chan) with structure of arrays (FASTEST)\n");
		float lambda=(float) 10.0;
		cudaTVpdhg_SoA(input,iter,nx,ny,lambda);
	}
}
